class YtdlError(Exception):
    pass


class NotFoundError(YtdlError):
    pass


class MultipleFilesError(YtdlError):
    pass


class MissingInfoError(YtdlError):
    pass


class AlreadyDownloadedError(YtdlError):
    pass



"""A Royal Games Wiki viewer :py:class:`royalnet.web.Royalprint`. Doesn't support any kind of edit."""

import flask as f
import markdown2
import re
import os
from ...royalprint import Royalprint
from ...shortcuts import error, from_urluuid
from ....database.tables import Royal, WikiPage, WikiRevision


tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
rp = Royalprint("wikiview", __name__, url_prefix="/wiki/view", template_folder=tmpl_dir,
                required_tables={Royal, WikiPage, WikiRevision})


class WikiRenderError(Exception):
    """An error occurred while trying to render the wiki page."""


def prepare_page_markdown(page):
    if list(page.content).count(">") > 99:
        raise WikiRenderError("Too many nested quotes")
    converted_md = markdown2.markdown(page.content.replace("<", "&lt;"),
                                      extras=["spoiler", "tables", "smarty-pants", "fenced-code-blocks"])
    converted_md = re.sub(r"{https?://(?:www\.)?(?:youtube\.com/watch\?.*?&?v=|youtu.be/)([0-9A-Za-z-]+).*?}",
                          r'<div class="youtube-embed">'
                          r'   <iframe src="https://www.youtube-nocookie.com/embed/\1?rel=0&amp;showinfo=0"'
                          r'           frameborder="0"'
                          r'           allow="autoplay; encrypted-media"'
                          r'           allowfullscreen'
                          r'           width="640px"'
                          r'           height="320px">'
                          r'   </iframe>'
                          r'</div>', converted_md)
    converted_md = re.sub(r"{https?://clyp.it/([a-z0-9]+)}",
                          r'<div class="clyp-embed">'
                          r'    <iframe width="100%" height="160" src="https://clyp.it/\1/widget" frameborder="0">'
                          r'    </iframe>'
                          r'</div>', converted_md)
    return f.Markup(converted_md)


def prepare_page(page):
    try:
        if page.format == "markdown":
            return f.render_template("wikiview_page.html",
                                     page=page,
                                     parsed_content=f.Markup(prepare_page_markdown(page)),
                                     css=page.css)
        elif page.format == "html":
            return f.render_template("wikiview_page.html",
                                     page=page,
                                     parsed_content=f.Markup(page.content),
                                     css=page.css)
        else:
            return error(500, f"Non esiste nessun handler in grado di preparare pagine con il formato {page.format}.")
    except WikiRenderError as e:
        return error(500, f"La pagina Wiki non può essere preparata a causa di un errore: {str(e)}")


@rp.route("/")
def wikiview_index():
    alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
    pages = sorted(alchemy_session.query(alchemy.WikiPage).all(), key=lambda page: page.title)
    return f.render_template("wikiview_index.html", pages=pages)


@rp.route("/<page_id>", defaults={"title": ""})
@rp.route("/<page_id>/<title>")
def wikiview_by_id(page_id: str, title: str):
    page_uuid = from_urluuid(page_id)
    alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
    page = alchemy_session.query(alchemy.WikiPage).filter(alchemy.WikiPage.page_id == page_uuid).one_or_none()
    if page is None:
        return error(404, f"La pagina richiesta non esiste.")
    return prepare_page(page)



"""Quick docs link :py:class:`royalnet.web.Royalprint`."""
import flask as f
from ...royalprint import Royalprint


rp = Royalprint("docs", __name__, url_prefix="/docs")


@rp.route("/")
def home_index():
    return f.redirect("https://royal-games.github.io/royalnet/html/index.html")



import typing
import asyncio
import pickle
from ..utils import Command, Call, NetworkHandler, asyncify
from ..network import Request, ResponseSuccess
from ..error import TooManyFoundError, NoneFoundError
from ..audio import YtdlDiscord
if typing.TYPE_CHECKING:
    from ..bots import DiscordBot


ytdl_args = {
    "format": "bestaudio",
    "outtmpl": f"./downloads/%(epoch)s_%(title)s.%(ext)s"
}


class PlayNH(NetworkHandler):
    message_type = "music_play"

    @classmethod
    async def discord(cls, bot: "DiscordBot", data: dict):
        """Handle a play Royalnet request. That is, add audio to a PlayMode."""
        # Find the matching guild
        if data["guild_name"]:
            guild = bot.client.find_guild(data["guild_name"])
        else:
            if len(bot.music_data) == 0:
                raise NoneFoundError("No voice clients active")
            if len(bot.music_data) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.music_data)[0]
        # Ensure the guild has a PlayMode before adding the file to it
        if not bot.music_data.get(guild):
            # TODO: change Exception
            raise Exception("No music_data for this guild")
        # Start downloading
        if data["url"].startswith("http://") or data["url"].startswith("https://"):
            dfiles: typing.List[YtdlDiscord] = await asyncify(YtdlDiscord.create_and_ready_from_url, data["url"], **ytdl_args)
        else:
            dfiles = await asyncify(YtdlDiscord.create_and_ready_from_url, f"ytsearch:{data['url']}", **ytdl_args)
        await bot.add_to_music_data(dfiles, guild)
        # Create response dictionary
        response = {
            "videos": [{
                "title": dfile.info.title,
                "discord_embed_pickle": str(pickle.dumps(dfile.info.to_discord_embed()))
            } for dfile in dfiles]
        }
        return ResponseSuccess(response)


async def notify_on_timeout(call: Call, url: str, time: float, repeat: bool = False):
    """Send a message after a while to let the user know that the bot is still downloading the files and hasn't crashed."""
    while True:
        await asyncio.sleep(time)
        await call.reply(f"ℹ️ Il download di [c]{url}[/c] sta richiedendo più tempo del solito, ma è ancora in corso!")
        if not repeat:
            break


class PlayCommand(Command):
    command_name = "play"
    command_description = "Riproduce una canzone in chat vocale."
    command_syntax = "[ [guild] ] (url)"

    network_handlers = [PlayNH]

    @classmethod
    async def common(cls, call: Call):
        guild_name, url = call.args.match(r"(?:\[(.+)])?\s*<?(.+)>?")
        download_task = call.loop.create_task(call.net_request(Request("music_play", {"url": url, "guild_name": guild_name}), "discord"))
        notify_task = call.loop.create_task(notify_on_timeout(call, url, time=30, repeat=True))
        try:
            data: dict = await download_task
        finally:
            notify_task.cancel()
        for video in data["videos"]:
            if call.interface_name == "discord":
                # This is one of the unsafest things ever
                embed = pickle.loads(eval(video["discord_embed_pickle"]))
                await call.channel.send(content="✅ Aggiunto alla coda:", embed=embed)
            else:
                await call.reply(f"✅ [i]{video['title']}[/i] scaricato e aggiunto alla coda.")



from sqlalchemy import Column, \
                       Integer, \
                       String
from sqlalchemy.ext.declarative import declared_attr
# noinspection PyUnresolvedReferences
from .keygroups import Keygroup


class Medal:
    __tablename__ = "medals"

    @declared_attr
    def mid(self):
        return Column(Integer, primary_key=True)

    @declared_attr
    def name(self):
        return Column(String, nullable=False)

    @declared_attr
    def description(self):
        return Column(String)

    @declared_attr
    def icon(self):
        return Column(String)

    @declared_attr
    def classes(self):
        return Column(String)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<Medal {self.mid}: {self.name}>"



from sqlalchemy import Column, \
                       Integer, \
                       DateTime, \
                       ForeignKey
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship
from .royals import Royal
from .medals import Medal


class MedalAward:
    __tablename__ = "MedalAward"

    @declared_attr
    def award_id(self):
        return Column(Integer, primary_key=True)

    @declared_attr
    def date(self):
        return Column(DateTime)

    @declared_attr
    def medal_id(self):
        return Column(Integer, ForeignKey("medals.mid"), nullable=False)

    @declared_attr
    def royal_id(self):
        return Column(Integer, ForeignKey("royal.uid"), nullable=False)

    @declared_attr
    def medal(self):
        return relationship("Medal", backref="awarded_to")

    @declared_attr
    def royal(self):
        return relationship("Royal", backref="medals_received")

    def __repr__(self):
        return f"<MedalAward of {self.medal} to {self.royal} on {self.date}>"



import contextlib
import os
import typing
import youtube_dl
from .ytdlinfo import YtdlInfo
from .errors import NotFoundError, MultipleFilesError, MissingInfoError, AlreadyDownloadedError


class YtdlFile:
    """Information about a youtube-dl downloaded file."""

    _default_ytdl_args = {
        "quiet": True,  # Do not print messages to stdout.
        "noplaylist": True,  # Download single video instead of a playlist if in doubt.
        "no_warnings": True,  # Do not print out anything for warnings.
        "outtmpl": "%(title)s-%(id)s.%(ext)s"  # Use the default outtmpl.
    }

    def __init__(self,
                 url: str,
                 info: typing.Optional[YtdlInfo] = None,
                 filename: typing.Optional[str] = None):
        self.url: str = url
        self.info: typing.Optional[YtdlInfo] = info
        self.filename: typing.Optional[str] = filename

    def has_info(self) -> bool:
        return self.info is not None

    def is_downloaded(self) -> bool:
        return self.filename is not None and os.path.exists(self.filename)

    @contextlib.contextmanager
    def open(self):
        if not self.is_downloaded():
            raise FileNotFoundError("The file hasn't been downloaded yet.")
        with open(self.filename, "r") as file:
            yield file

    def update_info(self, **ytdl_args) -> None:
        infos = YtdlInfo.retrieve_for_url(self.url, **ytdl_args)
        if len(infos) == 0:
            raise NotFoundError()
        elif len(infos) > 1:
            raise MultipleFilesError()
        self.info = infos[0]

    def download_file(self, **ytdl_args) -> None:
        if not self.has_info():
            raise MissingInfoError()
        if self.is_downloaded():
            raise AlreadyDownloadedError()
        with youtube_dl.YoutubeDL({**self._default_ytdl_args, **ytdl_args}) as ytdl:
            ytdl.download([self.info.webpage_url])
            self.filename = ytdl.prepare_filename(self.info.__dict__)

    def delete(self):
        if self.is_downloaded():
            os.remove(self.filename)
            self.filename = None

    @classmethod
    def download_from_url(cls, url: str, **ytdl_args) -> typing.List["YtdlFile"]:
        infos = YtdlInfo.retrieve_for_url(url, **ytdl_args)
        files = []
        for info in infos:
            file = YtdlFile(url=info.webpage_url, info=info)
            file.download_file(**ytdl_args)
            files.append(file)
        return files



import asyncio
from ..utils import Command, Call
from ..database.tables import Royal, Telegram, Discord


class RoyalnetprofileCommand(Command):

    command_name = "royalnetprofile"
    command_description = "Invia in chat il link al tuo profilo Royalnet!"
    command_syntax = ""

    require_alchemy_tables = {Royal, Telegram, Discord}

    @classmethod
    async def common(cls, call: Call):
        author = await call.get_author(error_if_none=True)
        if author is None:
            await call.reply("⚠️ Devi essere registrato a Royalnet per usare questo comando!")
            return
        await call.reply(f"🔗 https://ryg.steffo.eu/profile/{author.username}")



"""The profile page :py:class:`royalnet.web.Royalprint` for Royalnet members."""

import flask as f
import os
from ...royalprint import Royalprint
from ...shortcuts import error
from ....database.tables import *


# Maybe some of these tables are optional...
tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
rp = Royalprint("profile", __name__, url_prefix="/profile", template_folder=tmpl_dir,
                required_tables={Royal, ActiveKvGroup, Alias, Diario, Discord, Keygroup, Keyvalue, Telegram, WikiPage,
                                 WikiRevision})


@rp.route("/")
def profile_index():
    alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
    royals = alchemy_session.query(alchemy.Royal).order_by(alchemy.Royal.username).all()
    return f.render_template("profile_index.html", royals=royals)


@rp.route("/<username>")
def profile_by_username(username):
    alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
    royal = alchemy_session.query(alchemy.Royal).filter_by(username=username).one_or_none()
    if royal is None:
        return error(404, "Non esiste nessun utente con l'username richiesto.")
    return f.render_template("profile_page.html", royal=royal)



import discord


class FileAudioSource(discord.AudioSource):
    """A :py:class:`discord.AudioSource` that uses a :py:class:`io.BufferedIOBase` as an input instead of memory.

    The stream should be in the usual PCM encoding.

    Warning:
        This AudioSource will consume (and close) the passed stream."""

    def __init__(self, file):
        self.file = file

    def __repr__(self):
        if self.file.seekable():
            return f"<{self.__class__.__name__} @{self.file.tell()}>"
        else:
            return f"<{self.__class__.__name__}>"

    def is_opus(self):
        """This audio file isn't Opus-encoded, but PCM-encoded.

        Returns:
            ``False``."""
        return False

    def read(self):
        """Reads 20ms worth of audio.

        If the audio is complete, then returning an empty :py:class:`bytes`-like object to signal this is the way to do so."""
        # If the stream is closed, it should stop playing immediatly
        if self.file.closed:
            return b""
        data: bytes = self.file.read(discord.opus.Encoder.FRAME_SIZE)
        # If there is no more data to be streamed
        if len(data) != discord.opus.Encoder.FRAME_SIZE:
            # Close the file
            self.file.close()
            return b""
        return data



"""Video and audio downloading related classes, mainly used for Discord voice bots."""

from . import playmodes
from .ytdlinfo import YtdlInfo
from .ytdlfile import YtdlFile
from .fileaudiosource import FileAudioSource
from .ytdldiscord import YtdlDiscord

__all__ = ["playmodes", "YtdlInfo", "YtdlFile", "FileAudioSource", "YtdlDiscord"]



import telegram
import discord
import typing
from ..utils import Command, Call

class IdCommand(Command):

    command_name = "id"
    command_description = "Visualizza l'id della chat attuale."
    command_syntax = ""

    @classmethod
    def telegram(cls, call: Call):
        chat: telegram.Chat = call.channel
        call.reply(f"🔢 L'id di questa chat è [b]{chat.id}[/b].")

    @classmethod
    def discord(cls, call: Call):
        channel = call.channel
        if isinstance(channel, discord.TextChannel):
            call.reply(f"🔢 L'id di questa chat è [b]{channel.id}[/b].")
        elif isinstance(channel, discord.DMChannel):
            call.reply(f"🔢 L'id di questa chat è [b]{channel.id}[/b].")
        else:
            call.reply(f"⚠️ Questo tipo di chat non è supportato.")
